from io import StringIO
import unittest
from unittest.mock import ANY, patch

from muspinsim.input.gipaw import GIPAWOutput
from muspinsim.input.structure import MuonatedStructure
from muspinsim.tools.generator import (
    DipoleIntGenerator,
    GeneratorToolParams,
    QuadrupoleIntGenerator,
    QuadrupoleIntGeneratorGIPAWOut,
    _run_generator_tool,
    _select_atoms,
    generate_input_file_from_selection,
)

TEST_CELL_FILE_DATA = """%BLOCK LATTICE_CART
   14.18160000000000   0.000000000000000   0.000000000000000
   0.000000000000000   14.18160000000000   0.000000000000000
   0.000000000000000   0.000000000000000   14.18160000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
V             0.0791557133        0.0000005853        0.1708904764
V             0.9161667724        0.0000005823        0.8325662729
Si            0.4996287249        0.8331358706        0.8331663251
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC
"""

TEST_GIPAW_FILE_DATA = """     ----- total EFG -----
     V    1       -0.008877        0.000022       -0.011811
     V    1        0.000022       -0.030585        0.000005
     V    1       -0.011811        0.000005        0.039462

     V    2        0.221597       -0.000002       -0.009013
     V    2       -0.000002       -0.109627       -0.000009
     V    2       -0.009013       -0.000009       -0.111970

     Si   3       -0.002604       -0.000000       -0.000001
     Si   3       -0.000000       -0.002551       -0.000009
     Si   3       -0.000001       -0.000009        0.005154

     H    4       -0.034266        0.000000        0.000000
     H    4        0.000000       -0.034268        0.000000
     H    4        0.000000        0.000000        0.068534
"""

TEST_MAGRES_FILE_DATA = (
    "#$magres-abinitio-v1.0\n"
    "[atoms]\n"
    "units lattice Angstrom\n"
    "lattice          14.18160000000000   0.000000000000000   0.000000000000000"
    "                 0.000000000000000   14.18160000000000   0.000000000000000"
    "                 0.000000000000000   0.000000000000000   14.18160000000000\n"
    "units atom Angstrom\n"
    "atom V       V   1"
    "     1.1225546637        0.0000083005        2.4235003801\n"
    "atom V       V   2"
    "     12.992710700        0.0000082580        11.807121856\n"
    "atom Si      Si  3"
    "     7.0855347250        11.815199663        11.815631556\n"
    "atom H:mu    H:mu   1"
    "     2.3636086201        0.0000259155        1.1817982368\n"
    "[/atoms]\n"
    "[magres]\n"
    "units efg au\n"
    "efg V  1         -0.008877        0.000022       -0.011811"
    "                  0.000022       -0.030585        0.000005"
    "                 -0.011811        0.000005        0.039462\n"
    "efg V  2          0.221597       -0.000002       -0.009013"
    "                 -0.000002       -0.109627       -0.000009"
    "                 -0.009013       -0.000009       -0.111970\n"
    "efg Si 3         -0.002604       -0.000000       -0.000001"
    "                 -0.000000       -0.002551       -0.000009"
    "                 -0.000001       -0.000009        0.005154\n"
    "efg H:mu  1      -0.034266        0.000000        0.000000"
    "                  0.000000       -0.034268        0.000000"
    "                  0.000000        0.000000        0.068534\n"
    "[/magres]\n"
)


class TestGenerator(unittest.TestCase):
    def test_generate_input_file_cell(self):
        # No ignored symbols
        generate_params = GeneratorToolParams(
            structure=MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell"
            ),
            generators=[
                DipoleIntGenerator(),
                QuadrupoleIntGeneratorGIPAWOut(
                    GIPAWOutput(StringIO(TEST_GIPAW_FILE_DATA))
                ),
            ],
            number_closest=4,
            include_interatomic=False,
            additional_ignored_symbols=[],
        )

        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V Si Si
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539563139198 1.7614965359999998e-05 -1.2417021305964
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979205813603 1.7657510159999996e-05 3.5562763937592
quadrupolar 4
    -0.002604 -0.0 -1e-06
    -0.0 -0.002551 -9e-06
    -1e-06 -9e-06 0.005154
dipolar 1 4
    -4.72192610499264 2.366426252954879 3.54776669347968
quadrupolar 5
    -0.002604 -0.0 -1e-06
    -0.0 -0.002551 -9e-06
    -1e-06 -9e-06 0.005154
dipolar 1 5
    9.45967389500736 2.366426252954879 3.54776669347968
""",
        )

        # Ignore Si
        generate_params.additional_ignored_symbols = ["Si"]

        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V V V
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539563139198 1.7614965359999998e-05 -1.2417021305964
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979205813603 1.7657510159999996e-05 3.5562763937592
quadrupolar 4
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 4
    3.5524979205813603 1.7657510159999996e-05 -10.6253236062408
quadrupolar 5
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 5
    -10.62910207941864 1.7657510159999996e-05 3.5562763937592
""",
        )

        # Include interatomic interactions
        generate_params.include_interatomic = True
        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V V V
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539563139198 1.7614965359999998e-05 -1.2417021305964
dipolar 2 3
    2.3114439642674407 4.25447999999999e-08 4.7979785243555995
dipolar 2 4
    2.3114439642674407 4.25447999999999e-08 -9.3836214756444
dipolar 2 5
    -11.87015603573256 4.25447999999999e-08 4.7979785243555995
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979205813603 1.7657510159999996e-05 3.5562763937592
dipolar 3 4
    0.0 0.0 -14.1816
dipolar 3 5
    -14.1816 0.0 0.0
quadrupolar 4
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 4
    3.5524979205813603 1.7657510159999996e-05 -10.6253236062408
dipolar 4 5
    -14.1816 0.0 14.1816
quadrupolar 5
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 5
    -10.62910207941864 1.7657510159999996e-05 3.5562763937592
""",
        )

    def test_generate_input_file_magres(self):
        # No ignored symbols
        structure = MuonatedStructure(StringIO(TEST_MAGRES_FILE_DATA), fmt="magres")
        generate_params = GeneratorToolParams(
            structure=structure,
            generators=[
                DipoleIntGenerator(),
                QuadrupoleIntGenerator(structure),
            ],
            number_closest=4,
            include_interatomic=False,
            additional_ignored_symbols=[],
        )

        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V Si Si
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539564 1.7615e-05 -1.2417021433000002
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979200999995 1.7657499999999998e-05 3.556276380799999
quadrupolar 4
    -0.002604 -0.0 -1e-06
    -0.0 -0.002551 -9e-06
    -1e-06 -9e-06 0.005154
dipolar 1 4
    -4.7219261049 2.3664262525 3.5477666807999997
quadrupolar 5
    -0.002604 -0.0 -1e-06
    -0.0 -0.002551 -9e-06
    -1e-06 -9e-06 0.005154
dipolar 1 5
    9.4596738951 2.3664262525 3.5477666807999997
""",
        )

        # Ignore Si
        generate_params.additional_ignored_symbols = ["Si"]

        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V V V
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539564 1.7615e-05 -1.2417021433000002
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979200999995 1.7657499999999998e-05 3.556276380799999
quadrupolar 4
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 4
    3.5524979200999995 1.7657499999999998e-05 -10.6253236192
quadrupolar 5
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 5
    -10.6291020799 1.7657499999999998e-05 3.556276380799999
""",
        )

        # Include interatomic interactions
        generate_params.include_interatomic = True
        input_file_text = generate_input_file_from_selection(
            generate_params, _select_atoms(generate_params)
        )
        self.assertEqual(
            input_file_text,
            """spins
    mu V V V V
quadrupolar 2
    -0.008877 2.2e-05 -0.011811
    2.2e-05 -0.030585 5e-06
    -0.011811 5e-06 0.039462
dipolar 1 2
    1.2410539564 1.7615e-05 -1.2417021433000002
dipolar 2 3
    2.3114439636999995 4.2500000000000017e-08 4.7979785240999995
dipolar 2 4
    2.3114439636999995 4.2500000000000017e-08 -9.3836214759
dipolar 2 5
    -11.8701560363 4.2500000000000017e-08 4.7979785240999995
quadrupolar 3
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 3
    3.5524979200999995 1.7657499999999998e-05 3.556276380799999
dipolar 3 4
    0.0 0.0 -14.1816
dipolar 3 5
    -14.1816 0.0 0.0
quadrupolar 4
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 4
    3.5524979200999995 1.7657499999999998e-05 -10.6253236192
dipolar 4 5
    -14.1816 0.0 14.1816
quadrupolar 5
    0.221597 -2e-06 -0.009013
    -2e-06 -0.109627 -9e-06
    -0.009013 -9e-06 -0.11197
dipolar 1 5
    -10.6291020799 1.7657499999999998e-05 3.556276380799999
""",
        )

    def test_generate_input_error(self):
        # Should fail as indices in GIPAW file wont align
        generate_params = GeneratorToolParams(
            structure=MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell"
            ),
            generators=[
                DipoleIntGenerator(),
                QuadrupoleIntGeneratorGIPAWOut(
                    GIPAWOutput(
                        StringIO(
                            """     ----- total EFG -----
     V    1       -0.008877        0.000022       -0.011811
     V    1        0.000022       -0.030585        0.000005
     V    1       -0.011811        0.000005        0.039462

     V    2        0.221597       -0.000002       -0.009013
     V    2       -0.000002       -0.109627       -0.000009
     V    2       -0.009013       -0.000009       -0.111970

     Si 168       -0.002604       -0.000000       -0.000001
     Si 168       -0.000000       -0.002551       -0.000009
     Si 168       -0.000001       -0.000009        0.005154

     H  217       -0.034266        0.000000        0.000000
     H  217        0.000000       -0.034268        0.000000
     H  217        0.000000        0.000000        0.068534
"""
                        )
                    )
                ),
            ],
            number_closest=4,
            include_interatomic=False,
            additional_ignored_symbols=[],
        )

        with self.assertRaises(ValueError):
            generate_input_file_from_selection(
                generate_params, _select_atoms(generate_params)
            )

    # Mock these so dont actually do anything, just want to check parameters
    # used properly
    @patch("muspinsim.tools.generator.MuonatedStructure")
    @patch("muspinsim.tools.generator.generate_input_file_from_selection")
    def test_generate_tool_basic(
        self, generate_input_file_from_selection, mock_MuonatedStructure
    ):
        _run_generator_tool(
            [
                "V3Si_SC.cell",
                "4",
                "--dipolar",
            ],
            _select_atoms,
        )
        mock_MuonatedStructure.assert_called_with("V3Si_SC.cell", muon_symbol="H")
        generate_input_file_from_selection.assert_called_with(
            GeneratorToolParams(
                structure=mock_MuonatedStructure.return_value,
                generators=[ANY],
                number_closest=4,
                include_interatomic=False,
                additional_ignored_symbols=[],
                max_layer=6,
            ),
            mock_MuonatedStructure.return_value.compute_closest.return_value,
        )

    @patch("muspinsim.tools.generator.GIPAWOutput")
    @patch("muspinsim.tools.generator.MuonatedStructure")
    @patch("muspinsim.tools.generator.generate_input_file_from_selection")
    def test_generate_tool_gipaw(
        self,
        mock_generate_input_file_from_selection,
        mock_MuonatedStructure,
        mock_GIPAWOutput,
    ):
        _run_generator_tool(
            [
                "V3Si_SC.cell",
                "6",
                "--dipolar",
                "--quadrupolar",
                "./efg_nospin.out",
                "--ignore_symbol",
                "Si",
                "--ignore_symbol",
                "F",
                "--max_layer",
                "3",
                "--muon_symbol",
                "Cu",
            ],
            _select_atoms,
        )
        mock_MuonatedStructure.assert_called_with("V3Si_SC.cell", muon_symbol="Cu")
        mock_GIPAWOutput.assert_called_with("./efg_nospin.out")
        mock_generate_input_file_from_selection.assert_called_with(
            GeneratorToolParams(
                structure=mock_MuonatedStructure.return_value,
                generators=[ANY, ANY],
                number_closest=6,
                include_interatomic=False,
                additional_ignored_symbols=["Si", "F"],
                max_layer=3,
            ),
            mock_MuonatedStructure.return_value.compute_closest.return_value,
        )

    @patch("muspinsim.tools.generator.GIPAWOutput")
    @patch("muspinsim.tools.generator.MuonatedStructure")
    @patch("muspinsim.tools.generator.generate_input_file_from_selection")
    def test_generate_tool_invalid_quadrupolar(
        self,
        mock_generate_input_file_from_selection,
        mock_MuonatedStructure,
        mock_GIPAWOutput,
    ):
        # Don't supply a GIPAW file path while not using a magres file with
        # EFG tensors
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        mock_MuonatedStructure.return_value = structure
        with self.assertRaises(ValueError):
            _run_generator_tool(
                [
                    "V3Si_SC.cell",
                    "6",
                    "--dipolar",
                    "--quadrupolar",
                ],
                _select_atoms,
            )
        mock_MuonatedStructure.assert_called_with("V3Si_SC.cell", muon_symbol="H")

    @patch("muspinsim.tools.generator.GIPAWOutput")
    @patch("muspinsim.tools.generator.MuonatedStructure")
    @patch("muspinsim.tools.generator.generate_input_file_from_selection")
    def test_generate_tool_magres(
        self,
        mock_generate_input_file_from_selection,
        mock_MuonatedStructure,
        mock_GIPAWOutput,
    ):
        # Don't supply a GIPAW file path while using a magres file with EFG
        # tensors
        structure = MuonatedStructure(StringIO(TEST_MAGRES_FILE_DATA), fmt="magres")
        mock_MuonatedStructure.return_value = structure
        _run_generator_tool(
            [
                "V3Si_SC.cell",
                "6",
                "--dipolar",
                "--quadrupolar",
            ],
            _select_atoms,
        )
        mock_MuonatedStructure.assert_called_with("V3Si_SC.cell", muon_symbol="H")
        mock_generate_input_file_from_selection.assert_called_with(
            GeneratorToolParams(
                structure=mock_MuonatedStructure.return_value,
                generators=[ANY, ANY],
                number_closest=6,
                include_interatomic=False,
                additional_ignored_symbols=[],
                max_layer=6,
            ),
            ANY,
        )

    @patch("muspinsim.tools.generator.MuonatedStructure")
    @patch("muspinsim.tools.generator.generate_input_file_from_selection")
    def test_generate_tool_include_interatomic(
        self, mock_generate_input_file_from_selection, mock_MuonatedStructure
    ):
        _run_generator_tool(
            ["V3Si_SC.cell", "4", "--dipolar", "--include_interatomic"], _select_atoms
        )
        mock_MuonatedStructure.assert_called_with("V3Si_SC.cell", muon_symbol="H")
        mock_generate_input_file_from_selection.assert_called_with(
            GeneratorToolParams(
                structure=mock_MuonatedStructure.return_value,
                generators=[ANY],
                number_closest=4,
                include_interatomic=True,
                additional_ignored_symbols=[],
                max_layer=6,
            ),
            mock_MuonatedStructure.return_value.compute_closest.return_value,
        )



import unittest
import numpy as np

from muspinsim.spinop import SpinOperator, DensityOperator
from muspinsim.spinsys import MuonSpinSystem, SpinSystem
from muspinsim.hamiltonian import Hamiltonian


class TestHamiltonian(unittest.TestCase):
    def test_creation(self):

        H = np.array([[1, 0], [0, -1]])
        H = Hamiltonian(H)

        self.assertEqual(H.dimension, (2,))

        # Error if not Hermitian
        with self.assertRaises(ValueError) as e:
            Hamiltonian(np.array([[1, 1], [0, 1]]))

    def test_diag(self):

        Sx = SpinOperator.from_axes()
        H = Hamiltonian(Sx.matrix)

        evals, evecs = H.diag()
        evecsT = np.array([[1.0, 1.0], [-1.0, 1.0]]) / 2**0.5

        self.assertTrue(np.all(evals == [-0.5, 0.5]))
        self.assertTrue(np.all(np.isclose(abs(np.dot(evecs, evecsT)), np.eye(2))))

        Hrot = H.basis_change(evecs)

        self.assertTrue(np.all(np.isclose(Hrot.matrix.toarray(), np.diag(evals))))

    def test_evolve(self):

        ssys = SpinSystem(["e"])
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        rho0 = DensityOperator.from_vectors()  # Start along z
        t = np.linspace(0, 1, 100)

        self.assertTrue(isinstance(H, Hamiltonian))

        evol = H.evolve(rho0, t, ssys.operator({0: "z"}))

        self.assertTrue(np.all(np.isclose(evol[:, 0], 0.5 * np.cos(2 * np.pi * t))))

    def test_fast_evolve(self):

        ssys = MuonSpinSystem(["mu", "e"])
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        t = np.linspace(0, 1, 100)

        self.assertTrue(isinstance(H, Hamiltonian))

        # Start along z
        evol = H.fast_evolve(ssys.sigma_mu([0, 0, 1]), t, 2)

        self.assertTrue(np.all(np.isclose(evol, 0.5 * np.cos(2 * np.pi * t))))

    def test_integrate(self):

        ssys = SpinSystem(["e"])
        ssys.add_linear_term(0, [1, 0, 0])  # Precession around x
        H = ssys.hamiltonian
        rho0 = DensityOperator.from_vectors()  # Start along z

        self.assertTrue(isinstance(H, Hamiltonian))

        avg = H.integrate_decaying(rho0, 1.0, ssys.operator({0: "z"}))

        self.assertTrue(np.isclose(avg[0], 0.5 / (1.0 + 4 * np.pi**2)))



from io import StringIO
import unittest
import numpy as np

from muspinsim.input.structure import CellAtom, MuonatedStructure

TEST_CELL_FILE_DATA = """%BLOCK LATTICE_CART
   14.18160000000000   0.000000000000000   0.000000000000000
   0.000000000000000   14.18160000000000   0.000000000000000
   0.000000000000000   0.000000000000000   14.18160000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
V             0.0791557133        0.0000005853        0.1708904764
V             0.9161667724        0.0000005823        0.8325662729
Si            0.4996287249        0.8331358706        0.8331663251
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC
"""

TEST_MAGRES_FILE_DATA = (
    "#$magres-abinitio-v1.0\n"
    "[atoms]\n"
    "units lattice Angstrom\n"
    "lattice          14.18160000000000   0.000000000000000   0.000000000000000"
    "                 0.000000000000000   14.18160000000000   0.000000000000000"
    "                 0.000000000000000   0.000000000000000   14.18160000000000\n"
    "units atom Angstrom\n"
    "atom V       V   1"
    "     1.1225546637        0.0000083005        2.4235003801\n"
    "atom V       V   2"
    "     12.992710700        0.0000082580        11.807121856\n"
    "atom Si      Si  3"
    "     7.0855347250        11.815199663        11.815631556\n"
    "atom H:mu    H:mu   1"
    "     2.3636086201        0.0000259155        1.1817982368\n"
    "[/atoms]\n"
    "[magres]\n"
    "units efg au\n"
    "efg V  1         -0.008877        0.000022       -0.011811"
    "                  0.000022       -0.030585        0.000005"
    "                 -0.011811        0.000005        0.039462\n"
    "efg V  2          0.221597       -0.000002       -0.009013"
    "                 -0.000002       -0.109627       -0.000009"
    "                 -0.009013       -0.000009       -0.111970\n"
    "efg Si 3         -0.002604       -0.000000       -0.000001"
    "                 -0.000000       -0.002551       -0.000009"
    "                 -0.000001       -0.000009        0.005154\n"
    "efg H:mu  1      -0.034266        0.000000        0.000000"
    "                  0.000000       -0.034268        0.000000"
    "                  0.000000        0.000000        0.068534\n"
    "[/magres]\n"
)


def _optional_float_close(value1, value2):
    # Compares two values that may be a float or none
    if value1 is None or value2 is None:
        return value1 == value2
    else:
        return np.isclose(value1, value2)


def _optional_array_close(value1, value2):
    # Compares two values that may be an array or none
    if value1 is None or value2 is None:
        return value1 == value2
    else:
        return np.allclose(value1, value2)


class CellAtomMatcher:
    expected: CellAtom

    def __init__(self, expected):
        self.expected = expected

    def __repr__(self):
        return repr(self.expected)

    def __eq__(self, other):
        return (
            self.expected.index == other.index
            and self.expected.symbol == other.symbol
            and _optional_array_close(self.expected.position, other.position)
            and _optional_float_close(
                self.expected.distance_from_muon, other.distance_from_muon
            )
        )


class TestStructure(unittest.TestCase):
    def test_parsing_cell(self):
        # CASTEP cell file
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        self.assertEqual(len(structure._cell_atoms), 4)
        self.assertEqual(structure._muon_index, 3)
        self.assertEqual(
            CellAtomMatcher(structure._cell_atoms[2]),
            CellAtom(
                index=3,
                symbol="Si",
                isotope=1,
                position=np.array([7.08553473, 11.81519966, 11.81563156]),
                distance_from_muon=None,
            ),
        )
        self.assertEqual(structure.has_efg_tensors, False)

    def test_parsing_magres(self):
        # Magres file
        structure = MuonatedStructure(StringIO(TEST_MAGRES_FILE_DATA), fmt="magres")

        self.assertEqual(len(structure._cell_atoms), 4)
        self.assertEqual(structure._muon_index, 3)
        self.assertEqual(
            CellAtomMatcher(structure._cell_atoms[2]),
            CellAtom(
                index=3,
                symbol="Si",
                isotope=1,
                position=np.array([7.08553473, 11.81519966, 11.81563156]),
                distance_from_muon=None,
            ),
        )
        self.assertEqual(structure.has_efg_tensors, True)
        self.assertTrue(
            np.allclose(
                structure.get_efg_tensor(3),
                [
                    [-2.604e-03, -0.000e00, -1.000e-06],
                    [-0.000e00, -2.551e-03, -9.000e-06],
                    [-1.000e-06, -9.000e-06, 5.154e-03],
                ],
            )
        )

    def test_invalid_vector_angels(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(
                    """%BLOCK LATTICE_CART
   14.18160000000000   14.18160000000000   0.000000000000000
   0.000000000000000   14.18160000000000   0.000000000000000
   0.000000000000000   0.000000000000000   14.18160000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC"""
                ),
                fmt="castep-cell",
            )

    def test_no_muon_found(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA),
                muon_symbol="Cu",
                fmt="castep-cell",
            )

    def test_multiple_muons_found(self):
        with self.assertRaises(ValueError):
            MuonatedStructure(
                StringIO(TEST_CELL_FILE_DATA),
                muon_symbol="V",
                fmt="castep-cell",
            )

    def test_compute_layer_offsets(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        offsets = structure._compute_layer_offsets(1)
        self.assertEqual(len(offsets), 26)
        self.assertTrue(
            np.allclose(
                offsets,
                [
                    [-14.1816, -14.1816, -14.1816],
                    [-14.1816, -14.1816, 0.0],
                    [-14.1816, -14.1816, 14.1816],
                    [-14.1816, 0.0, -14.1816],
                    [-14.1816, 0.0, 0.0],
                    [-14.1816, 0.0, 14.1816],
                    [-14.1816, 14.1816, -14.1816],
                    [-14.1816, 14.1816, 0.0],
                    [-14.1816, 14.1816, 14.1816],
                    [14.1816, -14.1816, -14.1816],
                    [14.1816, -14.1816, 0.0],
                    [14.1816, -14.1816, 14.1816],
                    [14.1816, 0.0, -14.1816],
                    [14.1816, 0.0, 0.0],
                    [14.1816, 0.0, 14.1816],
                    [14.1816, 14.1816, -14.1816],
                    [14.1816, 14.1816, 0.0],
                    [14.1816, 14.1816, 14.1816],
                    [0.0, -14.1816, -14.1816],
                    [0.0, -14.1816, 0.0],
                    [0.0, -14.1816, 14.1816],
                    [0.0, 14.1816, -14.1816],
                    [0.0, 14.1816, 0.0],
                    [0.0, 14.1816, 14.1816],
                    [0.0, 0.0, -14.1816],
                    [0.0, 0.0, 14.1816],
                ],
            )
        )

        # Check correct number for larger system
        offsets = structure._compute_layer_offsets(2)
        self.assertEqual(len(offsets), 98)

        # Check offsets are correct for a system with different cell lengths
        structure = MuonatedStructure(
            StringIO(
                """%BLOCK LATTICE_CART
   14.18160000000000   0.000000000000000   0.000000000000000
   0.000000000000000   20.00000000000000   0.000000000000000
   0.000000000000000   0.000000000000000   8.000000000000000
%ENDBLOCK LATTICE_CART

%BLOCK POSITIONS_FRAC
H             0.1666672745        0.0000018274        0.0833332099
%ENDBLOCK POSITIONS_FRAC"""
            ),
            fmt="castep-cell",
        )
        offsets = structure._compute_layer_offsets(1)
        self.assertTrue(
            np.allclose(
                offsets,
                np.array(
                    [
                        [-14.1816, -20.0, -8.0],
                        [-14.1816, -20.0, 0.0],
                        [-14.1816, -20.0, 8.0],
                        [-14.1816, 0.0, -8.0],
                        [-14.1816, 0.0, 0.0],
                        [-14.1816, 0.0, 8.0],
                        [-14.1816, 20.0, -8.0],
                        [-14.1816, 20.0, 0.0],
                        [-14.1816, 20.0, 8.0],
                        [14.1816, -20.0, -8.0],
                        [14.1816, -20.0, 0.0],
                        [14.1816, -20.0, 8.0],
                        [14.1816, 0.0, -8.0],
                        [14.1816, 0.0, 0.0],
                        [14.1816, 0.0, 8.0],
                        [14.1816, 20.0, -8.0],
                        [14.1816, 20.0, 0.0],
                        [14.1816, 20.0, 8.0],
                        [0.0, -20.0, -8.0],
                        [0.0, -20.0, 0.0],
                        [0.0, -20.0, 8.0],
                        [0.0, 20.0, -8.0],
                        [0.0, 20.0, 0.0],
                        [0.0, 20.0, 8.0],
                        [0.0, 0.0, -8.0],
                        [0.0, 0.0, 8.0],
                    ]
                ),
            )
        )

    def test_compute_layer(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        new_atoms = structure.compute_layer(1)

        self.assertEqual(len(new_atoms), 3 * 26)  # Copy everything but muon
        self.assertEqual(
            CellAtomMatcher(new_atoms[11]),
            CellAtom(
                index=3,
                symbol="Si",
                isotope=1,
                position=np.array([-7.09606527, 11.81519966, -2.36596844]),
                distance_from_muon=None,
            ),
        )

    def test_compute_layer_ignore(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        new_atoms = structure.compute_layer(1, ignored_symbols=["Si"])

        self.assertEqual(len(new_atoms), 2 * 26)  # Copy everything but muon and silicon
        self.assertEqual(
            CellAtomMatcher(new_atoms[11]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([-1.18888930e00, 8.25794568e-06, 2.59887219e01]),
                distance_from_muon=None,
            ),
        )

    def test_compute_closest(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        closest_atoms = structure.compute_closest(1)

        self.assertEqual(len(closest_atoms), 1)
        self.assertEqual(
            CellAtomMatcher(closest_atoms[0]),
            CellAtom(
                index=1,
                symbol="V",
                isotope=1,
                position=np.array([1.12255466e00, 8.30049048e-06, 2.42350038e00]),
                distance_from_muon=1.7555737250028431,
            ),
        )

        more_closest_atoms = structure.compute_closest(6)

        self.assertEqual(len(more_closest_atoms), 6)
        self.assertEqual(CellAtomMatcher(closest_atoms[0]), more_closest_atoms[0])
        self.assertEqual(
            CellAtomMatcher(more_closest_atoms[5]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([1.29927107e01, 8.25794568e-06, -2.37447814e00]),
                distance_from_muon=11.208251995910084,
            ),
        )

    def test_compute_closest_ignore_all(self):
        # Ignore the muon as well
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        with self.assertRaises(ValueError):
            structure.compute_closest(1, ignored_symbols=["V", "Si", "H"])

        # Ignore all but taking the muon as Si
        structure = MuonatedStructure(
            StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell", muon_symbol="Si"
        )
        with self.assertRaises(ValueError):
            structure.compute_closest(1, ignored_symbols=["V", "Si", "H"])

    def test_compute_closest_ignore_all_but_muon(self):
        # Ignore everything but the muon
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        with self.assertRaises(RuntimeError):
            structure.compute_closest(1, ignored_symbols=["V", "Si"])

    def test_compute_closest_ignore_muon(self):
        # Ignore just the muon
        structure = MuonatedStructure(
            StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell", muon_symbol="H"
        )
        with self.assertRaises(ValueError):
            structure.compute_closest(1, ignored_symbols=["H"])

    def test_compute_closest_max_layer(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")

        with self.assertRaises(RuntimeError):
            structure.compute_closest(6, max_layer=1)

    def test_move_atom(self):
        structure = MuonatedStructure(StringIO(TEST_CELL_FILE_DATA), fmt="castep-cell")
        closest_atoms = structure.compute_closest(2)

        # Before moving
        self.assertEqual(
            CellAtomMatcher(closest_atoms[0]),
            CellAtom(
                index=1,
                symbol="V",
                isotope=1,
                position=np.array([1.12255466e00, 8.30049048e-06, 2.42350038e00]),
                distance_from_muon=1.7555737250028431,
            ),
        )
        self.assertEqual(
            CellAtomMatcher(closest_atoms[1]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([-1.18888930e00, 8.25794568e-06, -2.37447814e00]),
                distance_from_muon=5.0266632336825205,
            ),
        )

        # After moving atom closer to muon
        structure.move_atom(structure.muon, closest_atoms[0], 1)
        self.assertEqual(
            CellAtomMatcher(closest_atoms[0]),
            CellAtom(
                index=1,
                symbol="V",
                isotope=1,
                position=np.array([1.65668647e00, 1.58817187e-05, 1.88908961e0]),
                distance_from_muon=1.0,
            ),
        )
        self.assertEqual(
            CellAtomMatcher(closest_atoms[1]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([-1.18888930e00, 8.25794568e-06, -2.37447814e00]),
                distance_from_muon=5.0266632336825205,
            ),
        )
        self.assertTrue(
            np.isclose(
                np.linalg.norm(closest_atoms[0].position - structure.muon.position), 1.0
            )
        )

        # Try moving the two non muon's closer to ensure the distance to the
        # muon is recalculated
        structure.move_atom(closest_atoms[0], closest_atoms[1], 1)
        self.assertEqual(
            CellAtomMatcher(closest_atoms[1]),
            CellAtom(
                index=2,
                symbol="V",
                isotope=1,
                position=np.array([1.10155453e00, 1.43944274e-05, 1.05732731e00]),
                distance_from_muon=1.2681772563431124,
            ),
        )
        self.assertTrue(
            np.isclose(
                np.linalg.norm(closest_atoms[1].position - closest_atoms[0].position),
                1.0,
            )
        )

        # Should raise error if trying to move the muon
        with self.assertRaises(NotImplementedError):
            structure.move_atom(closest_atoms[0], structure.muon, 1)



from dataclasses import dataclass
import logging
import math
from pathlib import PurePath
from typing import IO, List, Optional, Union
import numpy as np
import ase.io
from numpy.typing import ArrayLike

from muspinsim.constants import spin


@dataclass
class CellAtom:
    """
    Dataclass storing information about an atom in a cell structure
    """

    index: int  # Start from 1, unchanged when creating a supercell
    symbol: str
    isotope: Optional[int]
    position: ArrayLike
    distance_from_muon: Optional[float] = None


def _get_isotope(mass: float, default_mass: float) -> Optional[int]:
    """
    Helper function to determine the isotope by comparing the found mass to
    the default of the same element

    At the moment will not return anything other than None as there is nowhere
    to get the isotope masses from.
    """

    # Expect close to whole division if valid
    if not math.isclose(mass, default_mass):
        raise ValueError("Failed to identify isotope by the given masses")

    return None


class MuonatedStructure:
    """
    A class for handling data from a muonated structure file and computing
    distances between the atoms
    """

    # Loaded data
    _unit_lengths: ArrayLike
    _unit_angles: ArrayLike
    _cell_atoms: List[CellAtom]

    _muon_symbol: str
    _muon_index: int

    _symbols_zero_spin: List[str]

    # Parameter optionally loaded from certain files (in this case magres)
    _efg_tensors: Optional[List[ArrayLike]]

    def __init__(
        self,
        file_io: Union[str, PurePath, IO],
        fmt: Union[str, None] = None,
        muon_symbol: Union[str, int] = "H",
    ):
        """Load a muonated cell structure file

        Read in a cell structure from an opened file stream

        Arguments:
            file_io {TextIOBase} -- File path, or IO stream
            fmt {str} -- Format of the file/data. See ase documentation for
                         available ones. When None ase will attempt to
                         determine it.
            muon_symbol {str} -- Symbol that should represent a muon.

        Raises:
            ValueError -- If the structure file has unit cell vector angles
                          different from [90, 90, 90].
            ValueError -- If the structure file contains more than one muon
                          with the given symbol.
            ValueError -- If the structure file has no muon with the given
                          symbol.
        """

        self._cell_atoms = []
        self._muon_index = None
        self._muon_symbol = muon_symbol
        self._symbols_zero_spin = []

        self._efg_tensors = None

        # Load the atomic data from the file
        # NOTE: Calculator is loaded automatically - can't see a way to
        # disable but this will cause warnings when reading CASTEP files
        # without a CASTEP installation - these can be ignored
        loaded_atoms = ase.io.read(file_io, format=fmt)

        lengths_and_angles = loaded_atoms.get_cell_lengths_and_angles()
        self._unit_lengths = lengths_and_angles[:3]
        self._unit_angles = lengths_and_angles[3:]

        # ase has a make_supercell method, unfortunately this is not suitable
        # as need indices to remain same for linking to other files
        if (self._unit_angles != 90).any():
            raise ValueError(
                f"Structure file '{file_io}' has unsupported unit cell angles"
            )

        # Load optional data
        if loaded_atoms.has("efg"):
            # Load the efg tensors
            self._efg_tensors = loaded_atoms.get_array("efg")

        self._cell_atoms = [None] * len(loaded_atoms)

        # Store only needed data for calculations
        for i, loaded_atom in enumerate(loaded_atoms):

            # Determine isotope by looking at loaded mass
            default_mass = ase.data.atomic_masses[
                ase.data.atomic_numbers[loaded_atom.symbol]
            ]
            isotope = _get_isotope(loaded_atom.mass, default_mass)

            if loaded_atom.symbol == muon_symbol:
                if self._muon_index is None:
                    self._muon_index = i
                else:
                    raise ValueError(
                        f"Structure file '{file_io}' has more than one muon "
                        f"with symbol {muon_symbol}"
                    )

            self._cell_atoms[i] = CellAtom(
                i + 1, loaded_atom.symbol, isotope, loaded_atom.position
            )

            # Keep track of any atoms with zero spin (so can exclude later)
            if (
                loaded_atom.symbol not in self._symbols_zero_spin
                and spin(elem=loaded_atom.symbol, iso=isotope) == 0
            ):
                self._symbols_zero_spin.append(loaded_atom.symbol)

        if self._muon_index is None:
            raise ValueError(
                f"Structure file '{file_io}' has no muon with symbol " f"{muon_symbol}"
            )

    def expand(
        self, offsets: List[ArrayLike], ignored_symbols: Optional[List[str]] = None
    ) -> List[CellAtom]:
        """Expands the structure by duplicating unit cell atoms (ignoring the
           muon)

        Expands the structure by cloning atoms (excluding the muon) given
        offsets to to modify their positions by.

        Arguments:
            offsets {List[ArrayLike]} -- Offset for the new atoms that will be
                                         added. For each offset will duplicate
                                         the current structure.
        """
        new_atoms = []
        for offset in offsets:
            # Copy everything but the muon
            for i, atom in enumerate(self._cell_atoms):
                if i != self._muon_index and (
                    ignored_symbols is None or atom.symbol not in ignored_symbols
                ):
                    new_atoms.append(
                        CellAtom(
                            index=atom.index,
                            symbol=atom.symbol,
                            isotope=atom.isotope,
                            position=atom.position + offset,
                        )
                    )
        return new_atoms

    def _compute_layer_offsets(self, layer: int) -> List[ArrayLike]:
        """Computes offsets for expanding this structure outwards

        Arguments:
            layer {int} -- Index of layer e.g. if layer = 1, will produce 8
                           offsets along the x, y and z axes. layer = 2 will
                           then produce another 26 offsets for expanding
                           another layer outwards.
        Returns:
            offsets {List[ArrayLike]} -- List of computed offsets
        """

        # Compute all potential values for the x, y and z axes
        potential_values = [
            np.arange(
                -layer * self._unit_lengths[i],
                (layer + 1) * self._unit_lengths[i],
                self._unit_lengths[i],
            )
            for i in range(3)
        ]

        # Obtain the extremes and values in between for each axis
        extreme_values = [
            [potential_values[i][0], potential_values[i][-1]] for i in range(3)
        ]
        middle_values = [potential_values[i][1:-1:] for i in range(3)]

        # Compute combinations
        offsets = []

        # Hold x fixed at one of the extreme values and vary others
        for x_value in extreme_values[0]:
            for y_value in potential_values[1]:
                for z_value in potential_values[2]:
                    offsets.append([x_value, y_value, z_value])

        # Middle values for x
        for x_value in middle_values[0]:
            # Keep y at the extremes
            for y_value in extreme_values[1]:
                for z_value in potential_values[2]:
                    offsets.append([x_value, y_value, z_value])

            # Middle values for y, with z at extremes
            for y_value in middle_values[1]:
                for z_value in extreme_values[2]:
                    offsets.append([x_value, y_value, z_value])

        return offsets

    def compute_layer(
        self, layer: int, ignored_symbols: Optional[List[str]] = None
    ) -> List[CellAtom]:
        """Duplicates atoms from the loaded cell by expanding outwards.

        Expands the structure outwards by duplicating it for a given layer
        number.

        Arguments:
            layer {int} -- Index of layer e.g. if layer = 1, will expand
                           along the x, y and z axes so that returned atoms
                           will have been computed for the 26 surrounding
                           cells. layer = 2 will then add compute those in
                           the next layer, which will include 98 cells etc.
            ignored_symbols {List[str]} -- List of symbols to ignore. May be
                                           None.
        Returns:
            new_atoms {List[CellAtom]} -- Additional atoms with their new
                                          positions.
        """

        # Now expand for each computed offset
        return self.expand(self._compute_layer_offsets(layer), ignored_symbols)

    def _compute_distances(self, atoms: List[CellAtom]):
        """Computes the vectors and distances between the muon and every atom."""
        muon = self._cell_atoms[self._muon_index]

        for atom in atoms:
            atom.distance_from_muon = np.linalg.norm(muon.position - atom.position)

    def compute_closest(
        self,
        number: int,
        ignored_symbols: Optional[List[str]] = None,
        max_layer: int = 6,
    ) -> List[CellAtom]:
        """Returns atoms the closest 'number' atoms to the muon.

        Expands structure outwards and computes distances until we are sure
        we have found the desired number of closest atoms to the muon.

        Arguments:
            number {int} -- Number of closest atoms to return. If number > the
                            number of atoms will only return what was
                            available.
            ignored_symbols {List[str]} -- List of symbols to ignore. May be
                                           None.
            max_layer {int} -- Maximum layer to allow the expansion to (for
                               avoiding any accidental long compute times)

        Returns:
            List[CellAtom] -- List of the found closest atoms to the muon
                              (ignoring any in ignored_symbols)

        Raises:
            ValueError -- When the muon_symbol is given in ignored_symbols
            RuntimeError -- When the expansion tries to go beyond the maximum
                            limit given by max_layer.
            RuntimeError -- When no atoms are found
        """

        # Current list of atoms in expanded supercell
        atoms = self._cell_atoms
        if ignored_symbols:
            # Ensure we are not trying to ignore the muon as well (this will
            # already be ignored in expansions anyway)
            if self._muon_symbol in ignored_symbols:
                raise ValueError(f"Cannot ignore the muon_symbol: {self._muon_symbol}")

            atoms = list(filter(lambda atom: atom.symbol not in ignored_symbols, atoms))

        # Due to the above validation, we must at least have a muon in the
        # system
        if len(atoms) <= 1:
            raise RuntimeError(
                "Failed to find any closest atoms, are they all in ignored_symbols?"
            )

        # Compute the distances within the cell
        self._compute_distances(atoms)

        continue_expansion = True
        layer = 1

        logging.info(
            "Attempting to find the %s closest atoms to the muon in the structure",
            number,
        )
        if ignored_symbols:
            logging.info("Ignoring the symbols: %s", ", ".join(ignored_symbols))

        while continue_expansion:
            # Sort by distance and obtain furthest distance of the desired
            # atom
            atoms = sorted(atoms, key=lambda atom: atom.distance_from_muon)
            furthest_atom = atoms[min(number, len(atoms) - 1)]

            logging.info("Expanding structure to layer: %d", layer)

            # Compute another layer and find the closest
            new_atoms = self.compute_layer(layer, ignored_symbols)
            self._compute_distances(new_atoms)
            new_atoms = sorted(new_atoms, key=lambda atom: atom.distance_from_muon)

            closest_new_atom = new_atoms[0]

            # Check whether we have just found atoms closer than the furthest
            # previous one or we dont have enough atoms in the structure
            if (
                closest_new_atom.distance_from_muon < furthest_atom.distance_from_muon
                or number + 1 > len(atoms)
            ):
                # Need this expansion to be included (and run at least
                # once more to check if there are any more to include)
                atoms.extend(new_atoms)

                layer += 1

                # Avoid taking forever
                if layer > max_layer:
                    raise RuntimeError("Trying to expand past the max_layer value")
            else:
                # No more expansion needed
                continue_expansion = False

        # Ignore the muon and obtain only the atoms requested
        atoms = atoms[1 : (number + 1)]

        for atom in atoms:
            logging.info(
                "Found %s at a distance of %s from the muon",
                atom.symbol,
                atom.distance_from_muon,
            )

        return atoms

    @property
    def symbols_zero_spin(self) -> List[str]:
        """
        List of symbols that have zero spin
        """
        return self._symbols_zero_spin

    @property
    def has_efg_tensors(self) -> bool:
        """Returns whether we have found and loaded EFG tensors"""
        return self._efg_tensors is not None

    def get_efg_tensor(self, atom_index: int) -> ArrayLike:
        """Returns the EFG tensor for a given atom index

        Arguments:
            index {int} -- Index of the relevant atom (starting from 1)
        Returns:
            efg_tensor {ArrayLike}: EFG tensor for the atom at the given index
        """
        return self._efg_tensors[atom_index - 1]

    @property
    def muon(self) -> CellAtom:
        """Returns the muon object from this structure"""
        return self._cell_atoms[self._muon_index]

    def move_atom(self, atom1: CellAtom, atom2: CellAtom, new_distance: float):
        """Moves atom2 to be a specific distance away from atom1, while preserving
        the direction between them

        Useful for adjusting distances where DFT calculations may be
        underestimating distances between the muon and its nearest neighbours

        Arguments:
            atom1 {CellAtom} -- Atom to compute the current distance from
            atom2 {CellAtom} -- Atom that will be moved
            new_distance {float} -- New distance that should be between the
                                    atoms after moving

        Raised:
            NotImplementedError: If atom2 is the muon (All the other positions would
                        need recalculating otherwise)
        """
        if atom2 == self.muon:
            raise NotImplementedError("Moving the muon is not supported")

        vector_between = atom2.position - atom1.position
        distance = np.linalg.norm(vector_between)

        # old_pos + new_dist * direction
        new_pos = atom1.position + ((new_distance / distance) * vector_between)

        # Log what is happening so can keep track
        logging.info(
            "Moving %s from %s to %s, changing the distance from %s to %s",
            atom2.symbol,
            atom2.position,
            new_pos,
            distance,
            new_distance,
        )

        atom2.position = new_pos

        # Update the distance to the muon (calculate if haven't already)
        if atom1 == self.muon:
            atom2.distance_from_muon = new_distance
        else:
            atom2.distance_from_muon = np.linalg.norm(
                atom2.position - self.muon.position
            )



import unittest
from muspinsim.input.larkeval import LarkExpression, LarkExpressionError, lark_tokenize
import numpy as np


def _double(x):
    return x * 2


class TestLark(unittest.TestCase):
    def _larkepxr_parse(self, test):
        """Lark expression unit test"""
        eval_vars = {k: v for k, v in test["vars"].items() if v is not None}
        if not test["invalid"]:
            e1 = LarkExpression(
                test["expr"],
                functions=test["funcs"],
                variables=list(test["vars"].keys()),
            )
            self.assertTrue(e1.evaluate(**eval_vars) == test["out"])
        else:
            with self.assertRaises(LarkExpressionError) as err:
                e1 = LarkExpression(
                    test["expr"],
                    functions=test["funcs"],
                    variables=list(test["vars"].keys()),
                )
                e1.evaluate(**eval_vars)

            self.assertTrue(test["out"] in str(err.exception))

    def test_lark_expr(self):
        self._larkepxr_parse(
            {"expr": "3+2*6^2/9", "funcs": {}, "vars": {}, "invalid": False, "out": 11}
        )

    def test_lark_func(self):

        self._larkepxr_parse(
            {
                "expr": "double(2)",
                "funcs": {"double": _double},
                "vars": {},
                "invalid": False,
                "out": 4,
            }
        )

    def test_lark_expr_in_func(self):
        self._larkepxr_parse(
            {
                "expr": "double(3+2*6^2/9)",
                "funcs": {"double": _double},
                "vars": {},
                "invalid": False,
                "out": 22,
            }
        )

    def test_lark_multi_expr_in_func(self):
        self._larkepxr_parse(
            {
                "expr": "double(3+2* 6^2/9) - double(2 *5)",
                "funcs": {"double": _double},
                "vars": {},
                "invalid": False,
                "out": 2,
            }
        )

    def test_lark_invalid_func(self):
        self._larkepxr_parse(
            {
                "expr": "notafunc(123)",
                "funcs": {"double": _double},
                "vars": {},
                "invalid": True,
                "out": "Invalid function: 'notafunc()', "
                "valid functions are ['double()']",
            }
        )

    def test_lark_invalid_expr(self):
        self._larkepxr_parse(
            {
                "expr": "3*",
                "funcs": {},
                "vars": {},
                "invalid": True,
                "out": "Invalid characters in LarkExpression: Unexpected token",
            }
        )

    def test_lark_empty(self):
        self._larkepxr_parse(
            {
                "expr": "",
                "funcs": {},
                "vars": {},
                "invalid": True,
                "out": "Empty String",
            }
        )

    def test_lark_var(self):
        self._larkepxr_parse(
            {"expr": "x+1", "funcs": {}, "vars": {"x": 5}, "invalid": False, "out": 6}
        )

    def test_lark_func_with_var(self):
        self._larkepxr_parse(
            {
                "expr": "double(x)",
                "funcs": {"double": _double},
                "vars": {"x": 2},
                "invalid": False,
                "out": 4,
            }
        )

    def test_lark_multi_var(self):
        self._larkepxr_parse(
            {
                "expr": "x+y+1",
                "funcs": {},
                "vars": {"x": 2, "y": 5},
                "invalid": False,
                "out": 8,
            }
        )

    def test_lark_vars_value_missing(self):
        self._larkepxr_parse(
            {
                "expr": "x+1",
                "funcs": {},
                "vars": {"x": None},
                "invalid": True,
                "out": "Some necessary variable(s) {'x'} have not "
                "been defined when evaluating LarkExpression",
            }
        )

    def test_lark_invalid_var(self):
        self._larkepxr_parse(
            {
                "expr": "x+1",
                "funcs": {},
                "vars": {"y": 2},
                "invalid": True,
                "out": "Invalid variable/constant: 'x', "
                "valid variables/constants are ['y']",
            }
        )

    def test_lark_evaluate_with_invalid_var(self):
        with self.assertRaises(LarkExpressionError) as err:
            out = LarkExpression("x+1", variables="x")
            out.evaluate(x=1, y=1)
        self.assertEqual(
            str(err.exception),
            "Some invalid variable(s) {'y'} have been defined when "
            "evaluating LarkExpression",
        )

    def test_tokenization(self):
        tokens = lark_tokenize("3+0.4 2.3 sin(x) atan2(3, 4)")
        lark_tokens = [
            LarkExpression(
                tk, variables="x", functions={"sin": np.sin, "atan2": np.arctan2}
            )
            for tk in tokens
        ]

        self.assertEqual(len(tokens), 4)
        self.assertEqual(lark_tokens[0].evaluate(), 3.4)
        self.assertEqual(lark_tokens[1].evaluate(), 2.3)
        self.assertAlmostEqual(lark_tokens[2].evaluate(x=np.pi / 2.0), 1.0)
        self.assertAlmostEqual(lark_tokens[3].evaluate(), np.arctan2(3.0, 4.0))



"""
MuSpinSim

A software to simulate the quantum dynamics of muon spin systems

Author: Simone Sturniolo

Copyright 2022 Science and Technology Facilities Council
This software is distributed under the terms of the MIT License
Please refer to the file LICENSE for the text of the license

"""

from muspinsim.input import MuSpinInput
from muspinsim.simconfig import MuSpinConfig
from muspinsim.experiment import ExperimentRunner
from muspinsim.fitting import FittingRunner
from muspinsim.version import __version__



import unittest
import numpy as np
import scipy.constants as cnst
from io import StringIO

from muspinsim import constants
from muspinsim.hamiltonian import Hamiltonian
from muspinsim.experiment import ExperimentRunner
from muspinsim.input import MuSpinInput


class TestExperiment(unittest.TestCase):
    def test_create(self):

        gmu = constants.MU_GAMMA
        ge = constants.ELEC_GAMMA

        stest = StringIO(
            """
name
    test
spins
    mu
zeeman 1
    0 0 1
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)
        ertest.B = np.array([0, 0, 1.0])

        self.assertTrue((ertest.Hsys.matrix == np.diag([0.5 * gmu, -0.5 * gmu])).all())
        self.assertTrue(np.allclose(ertest.Hz.matrix.data, ertest.Hsys.matrix.data))

        stest = StringIO(
            """
name
    test
spins
    e mu
"""
        )

        ertest.B = [0, 0, 1.0]

        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        ertest.B = np.array([0, 0, 1.0])

        self.assertEqual(ertest._system.elec_indices, {0})
        self.assertEqual(ertest._system.muon_index, 1)

        self.assertTrue(
            np.all(
                ertest.Hz.matrix.toarray()
                == np.diag(
                    [
                        0.5 * (ge + gmu),
                        0.5 * (ge - gmu),
                        0.5 * (-ge + gmu),
                        0.5 * (-ge - gmu),
                    ]
                )
            )
        )
        self.assertTrue(isinstance(ertest.Hz, Hamiltonian))

        Sx_mu = ertest._system.operator({1: "x"})
        Sz_e = ertest._system.operator({0: "z"})

        self.assertAlmostEqual(ertest.rho0.expectation(Sx_mu), 0.5)
        self.assertAlmostEqual(ertest.rho0.expectation(Sz_e), 0.0)

    def test_rho0(self):

        stest = StringIO(
            """
spins
    e mu
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        rho0 = ertest.rho0

        self.assertTrue(
            np.all(
                np.isclose(
                    rho0.matrix.toarray(),
                    [
                        [0.25, 0.25, 0, 0],
                        [0.25, 0.25, 0, 0],
                        [0, 0, 0.25, 0.25],
                        [0, 0, 0.25, 0.25],
                    ],
                )
            )
        )

        gmu = constants.MU_GAMMA
        ge = constants.ELEC_GAMMA

        T = 100
        ertest.B = [0, 0, 2.0e-6 * cnst.k * T / (ge * cnst.h)]
        ertest.T = T
        ertest.p = [0, 0, 1.0]

        rho0 = ertest.rho0

        Z = np.exp([-1, 1])
        Z /= np.sum(Z)

        self.assertTrue(
            np.all(np.isclose(np.diag(rho0.matrix.toarray()), [Z[0], 0, Z[1], 0]))
        )

    def test_run(self):

        # Empty system
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()

        self.assertTrue(np.all(results == 0.5))

        # Simple system
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        tau = constants.MU_TAU

        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
y_axis
    integral
x_axis
    field
field
    0
    1
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()

        self.assertAlmostEqual(results[0], 0.5 / (1.0 + 4 * np.pi**2 * tau**2))

        # Same result as above should hold for an intrinsic field
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
y_axis
    integral
x_axis
    intrinsic_field
intrinsic_field
    0
    1
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()

        self.assertAlmostEqual(results[0], 0.5 / (1.0 + 4 * np.pi**2 * tau**2))

    def test_run_intrinsic_field(self):
        # Check results from rotating are different when using field or intrinsic_field
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
orientation
    zcw(1)
field
    10 0 0
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()

        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
orientation
    zcw(1)
intrinsic_field
    10 0 0
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results_intrinsic = ertest.run()

        np.testing.assert_raises(
            AssertionError,
            np.testing.assert_array_almost_equal,
            results,
            results_intrinsic,
        )

    def test_run_celio(self):

        # Empty system
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
celio
    10
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        with self.assertRaises(ValueError):
            results = ertest.run()

        # Simple system that should fail as has dissipation
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
celio
    10
dissipation 1
    0.5
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        with self.assertRaises(NotImplementedError):
            results = ertest.run()

        # Simple system
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
celio
    10
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        # Simple system using faster method - should fail as muon is not first
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
celio
    10 8
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        self.assertEqual(ertest.config.celio_k, 10)
        self.assertEqual(ertest.config.celio_averages, 8)

        # Check raises error when muon is not first
        with self.assertRaises(ValueError):
            results = ertest.run()

        # Simple system using faster method
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
zeeman 1
    0 0 1.0/muon_gyr
celio
    10 8
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        # This test is subject to randomness, but np.isclose appears to avoid
        # any issues
        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        # Simple system using faster method with a temperature - should fail
        # as need T -> inf
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
celio
    10 8
temperature
    1.0
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        with self.assertRaises(ValueError):
            results = ertest.run()

    # For testing the faster time evolution method is run
    # correctly in the right scenarios
    def test_run_fast(self):

        # System without muon first should revert to the
        # slower method
        stest = StringIO(
            """
spins
    e mu
time
    range(0, 10)
zeeman 2
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertFalse(ertest._T_inf_speedup)
        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        # System with the muon first should use the faster method
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
zeeman 1
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertTrue(ertest._T_inf_speedup)
        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        # System with the a low temperature but no field should use the fast
        # method
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
temperature
    1
zeeman 1
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertTrue(ertest._T_inf_speedup)
        self.assertTrue(np.all(np.isclose(results, 0.5 * np.cos(2 * np.pi * times))))

        # System with T -> inf but a zero field should still use the fast
        # method
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
field
    0.01
zeeman 1
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertTrue(ertest._T_inf_speedup)

        # System with a low temperature and a non-zero field should not use
        # the fast method
        stest = StringIO(
            """
spins
    mu e
time
    range(0, 10)
field
    0.01
temperature
    1
zeeman 1
    0 0 1.0/muon_gyr
"""
        )
        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        results = ertest.run()
        times = ertest.config.x_axis_values

        self.assertFalse(ertest._T_inf_speedup)

    def test_dissipation(self):

        # Simple system
        g = 1.0
        stest = StringIO(
            """
spins
    mu
dissipation 1
    {g}
""".format(
                g=g
            )
        )

        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        times = ertest.config.x_axis_values

        results = ertest.run()

        solx = np.real(0.5 * np.exp(-g * times))
        self.assertTrue(np.all(np.isclose(results, solx)))

        # Check for temperature equilibrium
        T = 0.1
        stest = StringIO(
            """
spins
    mu
dissipation 1
    {g}
temperature
    {T}
field
    -1.0
time
    0
    20
""".format(
                g=g, T=T
            )
        )

        itest = MuSpinInput(stest)
        ertest = ExperimentRunner(itest)

        times = ertest.config.x_axis_values

        beta = 1.0 / (cnst.k * T)
        Z = np.exp(-cnst.h * constants.MU_GAMMA * 1e6 * beta)

        # Just let it evolve, see the result at long times
        Sz = ertest.system.operator({0: "z"})

        # Ready the first config
        ertest.load_config(ertest.config[0])
        results = ertest.Htot.evolve(ertest.rho0, times, operators=[Sz])

        self.assertAlmostEqual(np.real(results[-1, 0]), 0.5 * (1 - Z) / (1 + Z))



from abc import ABC, abstractmethod
import argparse
from dataclasses import dataclass, field
import logging
import sys
from typing import Callable, List

from muspinsim.input.structure import CellAtom, MuonatedStructure
from muspinsim.input.gipaw import GIPAWOutput


class InteractionGenerator(ABC):
    """
    Abstract base class for handling the generation of muspinsim interaction
    config
    """

    def __init__(self, num_inputs: int):
        """
        Number of atom indices expected as input for this generator e.g. 1 for
        quadrupole or 2 for dipole
        """
        self.num_inputs = num_inputs

    @abstractmethod
    def gen_config(self, atom_file_indices: List[int], atoms: List[CellAtom]) -> str:
        """Should return the config required for an interaction between the muon
        and given atom using this generator

        Arguments:
            atom_file_indices {List[int]} -- List of file indices for the
                                             atoms given as input (starts at
                                             1)
            atoms {List[CellAtom]} -- Atoms to generate config for
        """


class DipoleIntGenerator(InteractionGenerator):
    """
    Generator for the dipole interaction
    """

    def __init__(self):
        super().__init__(num_inputs=2)

    def gen_config(self, atom_file_indices: List[int], atoms: List[CellAtom]) -> str:
        vector_between = atoms[0].position - atoms[1].position

        return f"""dipolar {atom_file_indices[0]} {atom_file_indices[1]}
    {" ".join(map(str, vector_between))}"""


class QuadrupoleIntGeneratorGIPAWOut(InteractionGenerator):
    """
    Generator for the quadrupole interaction (From a GIPAW output file)
    """

    _gipaw_file: GIPAWOutput

    def __init__(self, gipaw_file: GIPAWOutput):
        super().__init__(num_inputs=1)

        self._gipaw_file = gipaw_file

    def gen_config(self, atom_file_indices: List[int], atoms: List[CellAtom]) -> str:
        # Obtain corresponding EFG tensor
        gipaw_atom = self._gipaw_file.find_atom(atoms[0].index)

        if gipaw_atom is None:
            raise ValueError(
                f"""Unable to locate atom with index {atoms[0].index} in GIPAW
output file"""
            )
        efg_tensor = gipaw_atom.efg_tensor

        return f"""quadrupolar {atom_file_indices[0]}
    {' '.join(map(str, efg_tensor[0]))}
    {' '.join(map(str, efg_tensor[1]))}
    {' '.join(map(str, efg_tensor[2]))}"""


class QuadrupoleIntGenerator(InteractionGenerator):
    """
    Generator for the quadrupole interaction (From data loaded with the
    structure file)
    """

    _structure: MuonatedStructure

    def __init__(self, structure: MuonatedStructure):
        super().__init__(num_inputs=1)

        self._structure = structure

    def gen_config(self, atom_file_indices: List[int], atoms: List[CellAtom]) -> str:
        # Obtain corresponding EFG tensor
        efg_tensor = self._structure.get_efg_tensor(atoms[0].index)

        return f"""quadrupolar {atom_file_indices[0]}
    {' '.join(map(str, efg_tensor[0]))}
    {' '.join(map(str, efg_tensor[1]))}
    {' '.join(map(str, efg_tensor[2]))}"""


@dataclass
class GeneratorToolParams:
    """Structure for storing parameters for the interaction generator tool

    Arguments:
        structure {MuonatedStructure} -- Structure to compute terms from
        generators {List[InteractionGenerator]} -- List of generators for
                                                generating interaction terms
        number_closest {int} -- Number of closest atoms to the muon to include
        include_interatomic {bool} -- Whether to include interactions between
                                      the atoms themselves (only applied to
                                      generators taking 2 inputs)
        additional_ignored_symbols {List[str]} -- List of additional symbols
                                    to ignore when counting the closest atoms.
                                    All spin 0 isotopes will be ignored by
                                    default but currently this data is not
                                    reflected by soprano for Si.
        max_layer {int} -- Maximum layer to allow the expansion to (for
                           avoiding any long compute times)
    """

    structure: MuonatedStructure
    generators: List[InteractionGenerator]
    number_closest: int
    include_interatomic: bool
    additional_ignored_symbols: List[str] = field(default_factory=list)
    max_layer: int = 6


def generate_input_file_from_selection(
    params: GeneratorToolParams, selected_atoms: List[CellAtom]
) -> str:
    """Utility function for generating muspinsim from a list of selected atoms
    obtained from a structure

    Arguments:
        params {GeneratorToolParams} -- Parameters (see above for details)
        selected_atoms {List[CellAtom]} -- Selected atoms to be included in
                                           the generated config
    Returns:
        input_file_text {str}: Generated muspinsim input file text
    """

    selected_symbols = "mu"
    input_file_text = ""
    muon = params.structure.muon

    # Split up generators based on how many inputs they take
    single_input_generators = list(
        filter(lambda generator: generator.num_inputs == 1, params.generators)
    )
    double_input_generators = list(
        filter(lambda generator: generator.num_inputs == 2, params.generators)
    )

    for i, selected_atom1 in enumerate(selected_atoms):
        selected_atom1_file_index = i + 2

        # Include the symbol into the spin's section
        symbol = selected_atom1.symbol

        if selected_atom1.isotope is not None:
            symbol = f"{str(selected_atom1.isotope)}{symbol}"

        selected_symbols += f" {symbol}"

        # Append config for single generators (using currently selected atom)
        for generator in single_input_generators:
            input_file_text += f"""{
                generator.gen_config(
                    atom_file_indices=[selected_atom1_file_index],
                    atoms=[selected_atom1],
                )
            }\n"""

        # Apply double input generators (with muon as first input)
        for generator in double_input_generators:
            input_file_text += f"""{
                generator.gen_config(
                    atom_file_indices=[1, selected_atom1_file_index],
                    atoms=[muon, selected_atom1],
                )
            }\n"""

        if params.include_interatomic:
            # Apply double terms between the nuclei themselves
            for j, selected_atom2 in enumerate(selected_atoms[i + 1 :], start=i + 1):
                selected_atom2_file_index = j + 2

                input_file_text += f"""{
                    generator.gen_config(
                        atom_file_indices=[
                            selected_atom1_file_index,
                            selected_atom2_file_index
                        ],
                        atoms=[selected_atom1, selected_atom2],
                    )
                }\n"""

    input_file_text = f"""spins
    {selected_symbols}
{input_file_text}"""

    return input_file_text


def _select_atoms(params: GeneratorToolParams) -> List[CellAtom]:
    """Selects atoms from a structure file to be used for generating a
    muspinsim input config

    Will expand a given muonated structure until the specified number
    of nearest neighbours are found

    Arguments:
        params {GeneratorToolParams} -- Parameters (see above for details)

    Returns:
        selected_atoms {List[CellAtom]}: Selected list of atoms
    """

    return params.structure.compute_closest(
        number=params.number_closest,
        ignored_symbols=params.structure.symbols_zero_spin
        + params.additional_ignored_symbols,
        max_layer=params.max_layer,
    )


def _run_generator_tool(
    args: List[str], selection_function: Callable[[GeneratorToolParams], List[CellAtom]]
):
    """Parsed the generator tool arguments from a given list and then
    runs it using the given selection_function to locate the nearest
    neighbours

    Arguments:
        args {List[str]} -- Arguments from the command line
        selection_function {function} -- Function that takes the generator
                                         tool params and returns the selected
                                         atoms to include in the config
                                         produced
    """

    # Setup so we can see the log output
    logging.basicConfig(
        format="[%(levelname)s] [%(asctime)s] %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    parser = argparse.ArgumentParser(
        description="""Generate a MuSpinSim input file using a structure file
(e.g. .cell, .cif, .magres).""",
    )

    # Required arguments
    parser.add_argument("filepath", type=str, help="Structure file filepath")
    parser.add_argument(
        "number_closest",
        type=int,
        help="Number of nearest atoms to the muon stopping site to include",
    )
    # Optional arguments
    parser.add_argument(
        "--dipolar",
        dest="dipolar",
        action="store_true",
        help="Specify to include dipolar couplings.",
    )
    parser.add_argument(
        "--quadrupolar",
        dest="gipaw_filepath",
        type=str,
        nargs="?",
        const="",  # When specified gives "" unless given value, otherwise None
        action="store",
        help="""Specify to include quadrupolar couplings. If the given
structure file is not a Magres file with EFG data a file path to a GIPAW
output file will be required.""",
    )
    parser.add_argument(
        "--include_interatomic",
        dest="include_interatomic",
        action="store_true",
        help="Specify to include interactions between the atoms themselves.",
    )
    parser.add_argument(
        "--out",
        dest="output",
        type=str,
        help="Filepath for the output file to generate",
    )
    parser.add_argument(
        "--muon_symbol",
        dest="muon_symbol",
        type=str,
        help="Symbol used to identify the muon (default: H)",
        default="H",
    )
    parser.add_argument(
        "--ignore_symbol",
        dest="ignored_symbols",
        action="append",
        help="""Allows certain elements in the cell file to be ignored by
their symbol when finding the closest atoms.""",
        default=[],
    )
    parser.add_argument(
        "--max_layer",
        dest="max_layer",
        type=int,
        help="""Maximum number of layers the structure can be expanded to
before raising an error (default: 6). Larger values may be needed when
searching for many atoms.""",
        default=6,
    )

    args = parser.parse_args(args)

    # Load the structure file
    structure = MuonatedStructure(args.filepath, muon_symbol=args.muon_symbol)

    # Terms to generate
    generators = []

    if args.dipolar:
        generators.append(DipoleIntGenerator())
    if args.gipaw_filepath is not None:
        # Override if specified
        if args.gipaw_filepath != "":
            generators.append(
                QuadrupoleIntGeneratorGIPAWOut(GIPAWOutput(args.gipaw_filepath))
            )
        else:
            # Attempt to find in the structure file
            if not structure.has_efg_tensors:
                raise ValueError(
                    """No EFGs found in the structure file, and
no GIPAW output file given. Please specify either '--quadrupolar
GIPAW_OUTPUT_FILEPATH' or supply a Magres structure file with EFG's provided."""
                )

            generators.append(QuadrupoleIntGenerator(structure))

    generate_params = GeneratorToolParams(
        structure=structure,
        generators=generators,
        number_closest=args.number_closest,
        include_interatomic=args.include_interatomic,
        additional_ignored_symbols=args.ignored_symbols,
        max_layer=args.max_layer,
    )
    file_data = generate_input_file_from_selection(
        generate_params, selection_function(generate_params)
    )

    if args.output is not None:
        with open(args.output, "w", encoding="utf-8") as file:
            file.write(file_data)
    else:
        print(file_data)


def main():
    """Entrypoint for command line tool"""

    _run_generator_tool(sys.argv[1:], _select_atoms)



"""larkeval.py

Evaluation functions and classes based off the Lark grammar parser"""

from typing import Dict, List, Optional
from lark import Lark
from lark.exceptions import UnexpectedToken, UnexpectedInput

_expr_parser = Lark(
    """
    ?start: sum

    ?function: NAME "(" (sum ("," sum)*)? ")" -> fun

    ?sum: product
        | sum "+" product   -> add
        | sum "-" product   -> sub

    ?product: power
        | product "*" power  -> mul
        | product "/" power  -> div

    ?power: atom
        | power "^" atom -> pow

    ?atom: NUMBER           -> num
         | "-" atom         -> neg
         | NAME             -> var
         | STRING           -> str
         | "(" sum ")"
         | function

    %import common.CNAME -> NAME
    %import common.NUMBER
    %import common.WS_INLINE
    %import common.ESCAPED_STRING -> STRING

    %ignore WS_INLINE
""",
    parser="lalr",
)


def lark_tokenize(line):
    """Splint a line of space-separated valid expressions."""
    ls = line.split()
    tokens = []
    while len(ls) > 0:
        tk = ls.pop(0)
        valid = False
        while not valid:
            try:
                _expr_parser.parse(tk)
                valid = True
            except UnexpectedToken as exc:
                if len(ls) == 0:
                    raise RuntimeError(
                        f"Line can not be tokenized into valid expressions: {str(exc)}"
                    ) from exc
                tk = tk + ls.pop(0)
            except UnexpectedInput as exc:
                raise LarkExpressionError(f"Could not parse input: {str(exc)}") from exc
        tokens.append(tk)

    return tokens


class LarkExpressionError(Exception):
    pass


class LarkExpression:
    def __init__(
        self,
        source,
        variables: Optional[List[str]] = None,
        functions: Optional[Dict[str, callable]] = None,
    ):
        """Create an expression to parse and evaluate with the Lark grammar
        parser. Variables and functions will be accepted only if included
        in the admissible ones.

        Arguments:
            source {str} -- Source code of the expression

        Keyword Arguments
            variables {[str]} -- Names of acceptable variables
            functions {{str: callable}} -- Names and bodies of acceptable
                                           functions
        """

        if variables is None:
            variables = []
        if functions is None:
            functions = {}

        # Start by parsing the expression
        self._source = source
        # check if source empty
        if source == "":
            raise LarkExpressionError("Empty String")
        try:
            self._tree = _expr_parser.parse(source)
        except UnexpectedToken as exc:
            raise LarkExpressionError(
                f"Invalid characters in LarkExpression: {str(exc)}"
            ) from exc

        # Find the variables and the function calls
        found_vars, found_funcs = self._analyse_tree(self._tree)

        self._variables = set(found_vars)
        self._functions = set(found_funcs)
        self._all_variables = set(
            [variables] if isinstance(variables, str) else variables
        )

        # Check if they are valid
        for v in self._variables:
            if v not in self._all_variables:
                raise LarkExpressionError(
                    "Invalid variable/constant: '{0}', "
                    "valid variables/constants are ['{1}']".format(
                        v, "', '".join(self._all_variables)
                    )
                )

        for f in self._functions:
            if f not in functions.keys():
                raise LarkExpressionError(
                    "Invalid function: '{0}()', valid functions are ['{1}()']".format(
                        f, "()', '".join(functions.keys())
                    )
                )

        self._function_bodies = {fn: functions[fn] for fn in self._functions}

        self._store_eval = None
        if len(self._variables) == 0:
            # No need to wait
            self._store_eval = self.evaluate()

    def _analyse_tree(self, root):
        """Traverse the tree to look for variables and functions"""

        found_vars = []
        found_functions = []

        if hasattr(root, "data"):
            children = root.children
            if root.data == "var":
                found_vars = [root.children[0].value]
            elif root.data == "fun":
                found_functions = [root.children[0].value]
                children = root.children[1:]  # Skip the NAME
        else:
            return [], []

        for c in children:
            fv, ff = self._analyse_tree(c)
            found_vars += fv
            found_functions += ff

        return found_vars, found_functions

    def _evaluate_tree(self, root, variables=None):
        if variables is None:
            variables = []

        if hasattr(root, "data"):
            d = root.data
            evc = [self._evaluate_tree(c, variables) for c in root.children]
            if d == "add":
                return evc[0] + evc[1]
            elif d == "sub":
                return evc[0] - evc[1]
            elif d == "mul":
                return evc[0] * evc[1]
            elif d == "div":
                return evc[0] / evc[1]
            elif d == "pow":
                return evc[0] ** evc[1]
            elif d == "neg":
                return -evc[0]
            elif d == "num":
                return float(evc[0])
            elif d == "str":
                return str(evc[0][1:-1])
            elif d == "var":
                return variables[evc[0]]
            elif d == "fun":
                fname = evc[0]
                return self._function_bodies[fname](*evc[1:])
        else:
            return root.value

    @property
    def functions(self):
        return self._functions

    @property
    def variables(self):
        return self._variables

    def evaluate(self, **variables):
        """Evaluate the expression with a given set of updated variables.

        Keyword Arguments:
            All the variable names appearing in self.variables.

        Returns:
            result {any} -- Result of evaluating the expression.
        """

        vset = set(variables.keys())
        if len(self.variables - vset) > 0:
            raise LarkExpressionError(
                f"Some necessary variable(s) {self.variables - vset} have not "
                "been defined when evaluating LarkExpression"
            )
        elif len(vset - self._all_variables) > 0:
            raise LarkExpressionError(
                f"Some invalid variable(s) {vset - self._all_variables} have "
                "been defined when evaluating LarkExpression"
            )

        if self._store_eval is not None:
            return self._store_eval
        else:
            return self._evaluate_tree(self._tree, variables)



"""hamiltonian.py

A class describing a spin Hamiltonian with various terms
"""

import numpy as np

from scipy import sparse

from muspinsim.cython import parallel_fast_time_evolve
from muspinsim.spinop import SpinOperator, DensityOperator, Operator, Hermitian
from muspinsim.validation import (
    validate_evolve_params,
    validate_integrate_decaying_params,
    validate_times,
)


class Hamiltonian(Operator, Hermitian):
    def __init__(self, matrix, dim=None):
        """Create an Hamiltonian

        Create an Hamiltonian from a hermitian complex matrix

        Arguments:
            matrix {ndarray} -- Matrix representation of the Hamiltonian

        Raises:
            ValueError -- Matrix isn't square or hermitian
        """

        super(Hamiltonian, self).__init__(matrix, dim)

    @classmethod
    def from_spin_operator(self, spinop):
        return self(spinop.matrix, spinop.dimension)

    def evolve(self, rho0, times, operators=None):
        """Time evolution of a state under this Hamiltonian

        Perform an evolution of a state described by a DensityOperator under
        this Hamiltonian and return either a sequence of DensityOperators or
        a sequence of expectation values for given SpinOperators.

        Arguments:
            rho0 {DensityOperator} -- Initial state
            times {ndarray} -- Times to compute the evolution for, in microseconds

        Keyword Arguments:
            operators {[SpinOperator]} -- List of SpinOperators to compute the
                                          expectation values of at each step.
                                          If omitted, the states' density
                                          matrices will be returned instead
                                           (default: {[]})

        Returns:
            [DensityOperator | ndarray] -- DensityOperators or expectation values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of times or operators
            RuntimeError -- Hamiltonian is not hermitian
        """
        if operators is None:
            operators = []

        times = np.array(times)

        if isinstance(operators, SpinOperator):
            operators = [operators]

        validate_evolve_params(rho0, times, operators)

        # Diagonalize self
        evals, evecs = self.diag()

        # Turn the density matrix in the right basis
        dim = rho0.dimension
        rho0 = rho0.basis_change(evecs).matrix.toarray()

        # Same for operators
        operatorsT = np.array(
            [o.basis_change(evecs).matrix.T.toarray() for o in operators]
        )

        # Matrix of evolution operators
        ll = -2.0j * np.pi * (evals[:, None] - evals[None, :])

        def calc_single_rho(i):
            return np.exp(ll[None, :, :] * times[i, None, None]) * rho0[None, :, :]

        result = None
        if len(operators) > 0:
            # Actually compute expectation values one at a time
            for i in range(times.shape[0]):
                rho = calc_single_rho(i)

                # This element wise multiplication then sum gives the equivalent
                # as the trace of the matrix product (without the transpose) and
                # and is faster
                single_res = np.sum(
                    rho[0, None, :, :] * operatorsT[None, :, :, :], axis=(2, 3)
                )
                if result is None:
                    result = single_res
                else:
                    result = np.concatenate(([result, single_res]), axis=0)
        else:
            sceve = evecs.T.conj()
            for i in range(times.shape[0]):
                # Just return density matrices
                result = [
                    DensityOperator(calc_single_rho(i), dim).basis_change(sceve)
                    for i in range(times.shape[0])
                ]
        return result

    def integrate_decaying(self, rho0, tau, operators=None):
        """Integrate one or more expectation values in time with decay

        Perform an integral in time from 0 to +inf of an expectation value
        computed on an evolving state, times a decay factor exp(-t/tau).
        This can be done numerically using evolve, but here is executed with
        a single evaluation, making it a lot faster.

        Arguments:
            rho0 {DensityOperator} -- Initial state
            tau {float} -- Decay time, in microseconds

        Keyword Arguments:
            operators {list} -- Operators to compute the expectation values
                                of (default: {[]})

        Returns:
            ndarray -- List of integral values

        Raises:
            TypeError -- Invalid operators
            ValueError -- Invalid values of tau or operators
            RuntimeError -- Hamiltonian is not hermitian
        """
        if operators is None:
            operators = []

        if isinstance(operators, SpinOperator):
            operators = [operators]

        validate_integrate_decaying_params(rho0, tau, operators)

        # Diagonalize self
        evals, evecs = self.diag()

        # Turn the density matrix in the right basis
        rho0 = rho0.basis_change(evecs).matrix.toarray()

        ll = 2.0j * np.pi * (evals[:, None] - evals[None, :])

        # Integral operators
        intops = np.array(
            [
                (-o.basis_change(evecs).matrix.toarray() / (ll - 1.0 / tau)).T
                for o in operators
            ]
        )

        result = np.sum(rho0[None, :, :] * intops[:, :, :], axis=(1, 2))

        return result

    def fast_evolve(self, sigma_mu, times, other_dimension):
        """Compute time evolution of a muon polarisation state using this
           Hamiltonian

        Computes the evolution of a muon polarisation state under this
        Hamiltonian and returns a sequence of expectation values.

        The muon polarisation is assumed to be first in the system in this
        computation.

        Arguments:
            sigma_mu {ndarray} -- Linear combination of Pauli spin matrices in
                                  the direction of the muon
            times {ndarray} -- Times to compute the evolution for, in microseconds
            other_dimension {int} -- Combined dimension of all non-muons in the
                                     system

        Returns:
            [ndarray] -- Expectation values

        Raises:
            ValueError -- Invalid values of times
            RuntimeError -- Hamiltonian is not hermitian
        """

        times = np.array(times)

        validate_times(times)

        # Diagonalize self
        evals, evecs = self.diag()

        # Expand to correct size (Assumes the muon is the first element in
        # the system)
        # Note: May be able to adapt method from C++ implementation of Celio's
        # to avoid the use of kron here to reduce memory and speed up -
        # although limiting factor at the moment is still the eigenvalue
        # computation
        sigma_mu = sparse.kron(sigma_mu, sparse.identity(other_dimension, format="csr"))

        # Compute the value of R^dagger * sigma * R
        A = np.dot(evecs.T.conjugate(), np.dot(sigma_mu.toarray(), evecs))

        # Mod square
        A = np.power(np.abs(A), 2)
        W = 2 * np.pi * np.subtract.outer(evals, evals)

        results = parallel_fast_time_evolve(times, other_dimension, A, W)

        # Divide by 2 as by convention rest of muspinsim gives results between
        # 0.5 and -0.5
        return results * 0.5



from io import StringIO
import unittest
import numpy as np

from muspinsim.input.gipaw import GIPAWAtom, GIPAWOutput

TEST_GIPAW_FILE_DATA = """     H  217        0.000008       -0.000024        0.068534

     ----- total EFG -----
     V    1       -0.008877        0.000022       -0.011811
     V    1        0.000022       -0.030585        0.000005
     V    1       -0.011811        0.000005        0.039462

     V    2        0.221597       -0.000002       -0.009013
     V    2       -0.000002       -0.109627       -0.000009
     V    2       -0.009013       -0.000009       -0.111970

     Si 168       -0.002604       -0.000000       -0.000001
     Si 168       -0.000000       -0.002551       -0.000009
     Si 168       -0.000001       -0.000009        0.005154

     H  217       -0.034266        0.000000        0.000000
     H  217        0.000000       -0.034268        0.000000
     H  217        0.000000        0.000000        0.068534


     NQR/NMR SPECTROSCOPIC PARAMETERS:
"""


class GIPAWAtomMatcher:
    expected: GIPAWAtom

    def __init__(self, expected):
        self.expected = expected

    def __repr__(self):
        return repr(self.expected)

    def __eq__(self, other):
        return self.expected.index == other.index and np.allclose(
            self.expected.efg_tensor, other.efg_tensor
        )


class TestGIPAW(unittest.TestCase):
    def test_parsing(self):
        gipaw_out_file = GIPAWOutput(StringIO(TEST_GIPAW_FILE_DATA))

        self.assertEqual(len(gipaw_out_file._atoms), 4)
        self.assertEqual(
            GIPAWAtomMatcher(gipaw_out_file._atoms[2]),
            GIPAWAtom(
                index=168,
                efg_tensor=np.array(
                    [
                        [-2.604e-03, -0.000e00, -1.000e-06],
                        [-0.000e00, -2.551e-03, -9.000e-06],
                        [-1.000e-06, -9.000e-06, 5.154e-03],
                    ]
                ),
            ),
        )

    def test_load_error(self):
        with self.assertRaises(ValueError):
            GIPAWOutput(
                StringIO(
                    """some random text
                       with newlines"""
                )
            )

    def test_parsing_errors(self):
        with self.assertRaises(ValueError):
            GIPAWOutput(
                StringIO(
                    """     ----- total EFG -----
     V    1       -0.008877        0.000022       -0.011811
     V    2        0.000022       -0.030585        0.000005
     V    1       -0.011811        0.000005        0.039462"""
                )
            )

        with self.assertRaises(ValueError):
            GIPAWOutput(
                StringIO(
                    """     ----- total EFG -----
     V    1       -0.008877        0.000022       -0.011811
     Si   1        0.000022       -0.030585        0.000005
     V    1       -0.011811        0.000005        0.039462"""
                )
            )

    def test_find_atom(self):
        gipaw_out_file = GIPAWOutput(StringIO(TEST_GIPAW_FILE_DATA))

        # Atom that will be at the corresponding index
        self.assertEqual(
            GIPAWAtomMatcher(gipaw_out_file.find_atom(2)),
            GIPAWAtom(
                index=2,
                efg_tensor=np.array(
                    [
                        [2.21597e-01, -2.00000e-06, -9.01300e-03],
                        [-2.00000e-06, -1.09627e-01, -9.00000e-06],
                        [-9.01300e-03, -9.00000e-06, -1.11970e-01],
                    ]
                ),
            ),
        )

        # This will not be at the corresponding index
        self.assertEqual(
            GIPAWAtomMatcher(gipaw_out_file.find_atom(168)),
            GIPAWAtom(
                index=168,
                efg_tensor=np.array(
                    [
                        [-2.604e-03, -0.000e00, -1.000e-06],
                        [-0.000e00, -2.551e-03, -9.000e-06],
                        [-1.000e-06, -9.000e-06, 5.154e-03],
                    ]
                ),
            ),
        )
